"""Miscellaneous functions for Genbabel package.
"""

import datetime


def getfilename():
    """Return the filename based on the datetime.

    :return: the filename in year-month-day_hour-minute
    :rtype: str
    """

    timenow = datetime.datetime.now()

    year = str(timenow.year % 100)
    month = str(timenow.month).zfill(2)
    day = str(timenow.day).zfill(2)
    hour = str(timenow.hour).zfill(2)
    minute = str(timenow.minute).zfill(2)

    filename = year + month + day + "_" + hour + minute

    return filename



# pylint: disable=C0103,E0401
"""
Synbiopython (c) Global BioFoundry Alliance 2020

Synbiopython is licensed under the MIT License.

This module is to generate SEDML and COMBINE OMEX files.
"""

import os
import re
import tempfile
import io
import zipfile
import requests
import tellurium as te
import matplotlib.pyplot as plt
import phrasedml
from synbiopython.genbabel import utilities

te.setDefaultPlottingEngine("matplotlib")

plt.close("all")

plt.rcdefaults()
plt.rcParams["font.family"] = "Arial"
plt.rcParams["font.weight"] = "normal"
plt.rcParams["font.size"] = 16
plt.rcParams["axes.labelsize"] = 16
plt.rcParams["axes.labelweight"] = "normal"
plt.rcParams["axes.linewidth"] = 2
plt.rcParams["axes.formatter.limits"] = -3, 3
plt.rcParams["legend.frameon"] = False
params = {"mathtext.default": "regular"}
plt.rcParams.update(params)
plt.rcParams.update({"axes.spines.top": False, "axes.spines.right": False})


class SEDMLOMEXgen:
    """Class to generate the SEDML and COMBINE OMEX files."""

    # get current working directory
    def __init__(self):
        self.workingdir0 = os.getcwd()

    @staticmethod
    def get_sbml_biomodel(Biomodels_ID, **kwargs):
        """Get SBML model from biomodel. Use outputfile keyword argument to
        export the SBML model into .xml file at the specific path.
        A default .xml file will be generated by default at the temp directory.

        :param Biomodels_ID: the ID for the Biomodels
        :type Biomodels_ID: str
        :return: the sbml in string format
        :rtype: str
        :Keyword Arguments:
            * *outputfile*: specify outputfile
        """

        temppath = tempfile.gettempdir()

        url = "https://www.ebi.ac.uk/biomodels/search/download?models=" + Biomodels_ID
        r = requests.get(url, stream=True)
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(temppath)

        filepath = os.path.join(temppath, Biomodels_ID + ".xml")

        with open(filepath, "r") as f:
            biomodelsbml_str = f.read()

        for key, value in kwargs.items():
            if "outputfile" in key:
                filepath = value
                with open(filepath, "wb") as f:
                    f.write(biomodelsbml_str.encode("utf-8"))

        print("The SBML file path: ", filepath)

        return biomodelsbml_str

    @staticmethod
    def sbmltoantimony(sbmlfile):
        """Get the sbml file and return the antimony string."""

        antimony_str = te.sbmlToAntimony(sbmlfile)
        basename = os.path.basename(sbmlfile)
        sbmlfilename = basename.split(".")[0]
        antimony_str = antimony_str.replace("doc0", sbmlfilename)

        return antimony_str

    @staticmethod
    def find_between(s, first, last):
        """Get the substring from string based on indexes.

        :param s: string to be searched
        :type s: str
        :param first: part of the string at the front
        :type first: str
        :param last: part of the string at the end
        :type last: str
        """

        try:
            start = s.index(first) + len(first)
            end = s.index(last, start)
            return s[start:end]
        except ValueError:
            return ""

    def export_omex(self, antimony_str, phrasedml_str, **kwargs):
        """Generate COMBINE OMEX file.

        :param antimony_str: represent the SBML
        :type antimony_str: str
        :param phrasedml_str: represent the SEDML
        :type phrasedml_str: str
        :return: omex inline
        :rtype: str
        :Keyword Arguments:
            * *outputfile*: specify outputfile
        """

        model = re.search("model (.*)\n", antimony_str)

        if model.group(1)[0] == "*":
            model = self.find_between(model.group(1), "*", "()")
            phrasedml_str = phrasedml_str.format(model)
        else:
            phrasedml_str = phrasedml_str.format(model.group(1))

        inline_omex = "\n".join([antimony_str, phrasedml_str])

        if "outputfile" in kwargs:
            filepath = kwargs["outputfile"]

        else:
            dirName = self.get_omexfilename()

            try:
                os.mkdir(dirName)
                print("Directory ", dirName, " Created ")
            except FileExistsError:
                print("Directory ", dirName, " already exists")

            workingDir = os.path.join(self.workingdir0, dirName)

            filepath = os.path.join(workingDir, "archive.omex")

        print("The output file path: ", filepath)

        te.exportInlineOmex(inline_omex, filepath)

        return inline_omex

    @staticmethod
    def execute_inlineomex(inline_omex):
        """Execute the inline omex and generate the simulation figures.
        """

        return te.executeInlineOmex(inline_omex)

    def phrasedmltosedml(self, phrasedml_str, sbml_file, **kwargs):
        """Generate SEDML file from phrasedml.
        Example of phrasedml_str:
        phrasedml_str = '''
        model1 = model "{}"
        ...
        ...
        '''

        :param phrasedml_str: text-based way to represent SEDML
        :type phrasedml_str: str
        :param sbml_file: the SBML xml file/path to the file
        :type sbml_file: str
        :return: the sedml string
        :rtype: str
        """

        try:
            with open(sbml_file, "r+") as f:
                sbml_str = f.read()
        except IOError:
            print("Error in opening sbml file")

        phrasedml_str = phrasedml_str.format(sbml_file)

        phrasedml.setReferencedSBML(sbml_file, sbml_str)
        sedml_str = phrasedml.convertString(phrasedml_str)

        if sedml_str is None:
            raise RuntimeError(phrasedml.getLastError())

        sedml_file = os.path.join(self.workingdir0, "sedml.xml")

        for key, value in kwargs.items():
            if "outputfile" in key:
                sedml_file = value

        with open(sedml_file, "wb") as f:
            f.write(sedml_str.encode("utf-8"))

        return sedml_str

    @staticmethod
    def get_omexfilename():
        """Return filename to the OMEX file according to the export time."""

        omexfilename = "OMEX" + utilities.getfilename()

        return omexfilename



"""
Synbiopython (c) Global BioFoundry Alliance 2020

Synbiopython is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author: neilswainston
"""
import os.path

DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")



"""
Synbiopython (c) Global BioFoundry Alliance 2020

Synbiopython is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author: neilswainston
"""
from collections import defaultdict
import os.path
import re
from urllib.request import urlretrieve

from synbiopython.codon.taxonomy_utils import get_tax_id

_CODON_REGEX = r"([ATGCU]{3}) ([A-Z]|\*) (\d.\d+)"


def get_table(table_id, dna=True):
    """Gets a codon table from from supplied parameter, which may be either an
    organism name or a NCBI Taxonomy id.

    :param table_id: an organism name or a NCBI Taxonomy id (as either a str or
        int).
    :type table_id: str
    :param dna: boolean parameter specifying whether the codon table returned
        should contain DNA or RNA codons (default is DNA).
    :type dna: bool
    :return: a codon usage table.
    :rtype: dict
    """
    tax_id = get_tax_id(table_id)
    results = defaultdict(dict)
    content = _get_content(tax_id)

    for vals in sorted(re.findall(_CODON_REGEX, content),
                       key=lambda x: (x[1], x[2])):
        results[vals[1]][vals[0].replace("U", "T") if dna else vals[0]] = \
            float(vals[2])

    return dict(results)


def _get_content(tax_id):
    """Gets Codon Usage Database content, either from cached file or remotely.

    :param tax_id: a NCBI Taxonomy id.
    :type tax_id: str
    :return: the Codon Usage Database content
    :rtype: str
    """
    target_dir = os.path.join(os.path.expanduser('~'), '.synbiopython/codon')

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    target_file = os.path.join(target_dir, "%s.txt" % tax_id)

    if not os.path.exists(target_file):
        url = "http://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?" + \
            "aa=1&style=N&species=%s" % tax_id

        urlretrieve(url, target_file)

    with open(target_file) as fle:
        return fle.read()



"""Classes to represent plates"""
from synbiopython.lab_automation.containers.Plate import Plate
from synbiopython.lab_automation.containers.Well import Well


class Plate96(Plate):
    """Base class for standard 96-well plates"""

    num_rows = 8
    num_columns = 12


class Plate384(Plate):
    """Base class for standard 384-well plates"""

    num_rows = 16
    num_columns = 24


class Plate1536(Plate):
    """Base class for 1536-well plates (32 rows, 48 columns)"""

    num_rows = 32
    num_columns = 48


class Plate2x4(Plate):
    """Class for 8-well (2 x 4) plates such as colony plating plates"""

    num_rows = 2
    num_columns = 4


# Plate4ti0960


class Plate4ti0960Well(Well):
    """Well for 96-well plate from 4titude"""

    capacity = 150e-6


class Plate4ti0960(Plate96):
    """96-well plate from 4titude"""

    well_class = Plate4ti0960Well


# Plate4ti0130


class Plate4ti0130Well(Well):
    """Well for 96-well plate with 2ml deepwells from 4titude"""

    capacity = 1900e-6


class Plate4ti0130(Plate96):
    """96-well plate with 2ml deepwells from 4titude"""

    well_class = Plate4ti0130Well


# PlateLabcyteEchoLp0200Ldv


class PlateLabcyteEchoLp0200LdvWell(Well):
    """Well for low dead volume 384-well Echo plate"""

    capacity = 12e-6
    echo_dead_volume = 3e-6


class PlateLabcyteEchoLp0200Ldv(Plate384):
    """Low dead volume 384-well Echo plate"""

    class PlateWell(Well):
        """Well"""

        capacity = 12e-6
        echo_dead_volume = 3e-6


class PlateLabcyteEchoP05525Pp(Plate384):
    """Polypropylene 384-well ECHO plate"""

    class PlateWell(Well):
        """Well"""

        capacity = 50e-6
        echo_dead_volume = 15e-6


class Trough8x1(Plate):
    """Eight positions share the same content"""

    num_rows = 8
    num_columns = 1

    def __init__(self, name, wells_data=None, plate_data=None):
        Plate.__init__(self, name=name, wells_data=None, plate_data=None)
        for well in self:
            well.content = self["A1"].content



# pylint: disable=C0103,C0330,C0114,C0115,C0116,C0301


class TransferError(ValueError):
    pass


class Transfer:
    """Class representing a transfer from a source well to a destination well.

    :param source_well: A Well object from which to transfer.
    :param destination_well: A Well object to which to transfer.
    :param volume: Volume to be transferred, expressed in liters.
    :param data: A dict containing any useful information about the transfer.
        This information can be used later e.g. as parameters for the transfer
        when exporting a picklist.
    """

    def __init__(self, source_well, destination_well, volume, data=None):

        self.volume = volume
        self.source_well = source_well
        self.destination_well = destination_well
        self.data = data

    def to_plain_string(self):
        """Return "Transfer {volume}L from {source_well} into {dest_well}"."""
        return (
            "Transfer {self.volume:.02E}L from {self.source_well.plate.name} "
            "{self.source_well.name} into "
            "{self.destination_well.plate.name} "
            "{self.destination_well.name}"
        ).format(self=self)

    def to_short_string(self):
        """Return "Transfer {volume}L {source_well} -> {dest_well}"."""
        return (
            "{self.__class__.__name__} {self.volume:.02E}L {self.source_well} -> {self.destination_well}"
        ).format(self=self)

    def with_new_volume(self, new_volume):
        """Return a version of the transfer with a new volume."""
        return self.__class__(
            source_well=self.source_well,
            destination_well=self.destination_well,
            volume=new_volume,
            data=self.data,
        )

    def apply(self):
        # error_prefix = "%s error:" % self.to_short_string()

        if self.source_well.is_empty:
            raise TransferError("Source well is empty!")

        #  pre-check in both source and destination wells that transfers
        #  are valid
        if self.volume > self.source_well.volume:
            raise TransferError(
                ("Subtraction of %.2e L from %s impossible." " Current volume: %.2e L")
                % (self.volume, self, self.source_well.volume)
            )
        final_destination_volume = self.destination_well.volume + self.volume
        if (self.destination_well.capacity is not None) and (
            final_destination_volume > self.destination_well.capacity
        ):
            raise TransferError(
                "Transfer of %.2e L from %s to %s brings volume over capacity."
                % (self.volume, self, self.destination_well)
            )

        #  If you arrive here, it means that the transfer is valid, do it.
        factor = float(self.volume) / self.source_well.volume

        quantities_transferred = {
            component: quantity * factor
            for component, quantity in self.source_well.content.quantities.items()
        }
        self.destination_well.add_content(quantities_transferred, volume=self.volume)
        self.source_well.subtract_content(quantities_transferred, volume=self.volume)
        if self not in self.destination_well.sources:
            self.destination_well.sources.append(self)

    def __repr__(self):
        """Return  "Transfer {volume}L from {source_well} into {dest_well}"."""
        return self.to_plain_string()



# pylint: disable=C0114
from .containers.Plate import Plate
from .containers.builtin_containers import (
    Plate96,
    Plate384,
    Plate1536,
    Plate2x4,
    Plate4ti0960,
    Plate4ti0130,
    PlateLabcyteEchoLp0200Ldv,
    PlateLabcyteEchoP05525Pp,
    Trough8x1,
)
from .picklist.PickList import PickList, Transfer
from .picklist.Transfer import TransferError



"""
Synbiopython (c) Global BioFoundry Alliance 2020

Synbiopython is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author: neilswainston
"""
import random

_CODON_REGEX = r"([ATGCU]{3}) ([A-Z]|\*) (\d.\d+)"


def sample(table, amino_acid):
    """Sample a codon for a given amino acid probabilistically, based on its
    codon usage frequency.

    :param table: a codon usage table.
    :type table: dict
    :param amino_acid: a single-character string representing an amino acid.
    :type amino_acid: str
    :return: a codon encoding the supplied amino acid, sampled
        probabilistically.
    :rtype: str
    """
    cum_freq = 0
    rand_val = random.random()

    for codon, freq in table[amino_acid].items():
        cum_freq += freq

        if cum_freq > rand_val:
            return codon

    return None


def optimise(table, aa_seq):
    """Codon optimise an amino acid sequence.

    :param table: a codon usage table.
    :type table: dict
    :param aa_seq: an amino acid sequence.
    :type table: str
    :return: a codon-optimised nucleic acid sequence, encoding the supplied
        amino acid sequence
    :rtype: str
    """
    return ''.join([sample(table, amino_acid) for amino_acid in aa_seq])



# pylint: disable=C0330,C0103,C0116,R0201,R0912,R0913,R0914,R0915,R0904
"""
This module is the simplesbml package from https://simplesbml.readthedocs.io/en/latest/
with minor modifications to include more unit definitions

Reference:
    https://github.com/sys-bio/simplesbml/blob/master/simplesbml/__init__.py
"""

import warnings

# try to import tesbml or libsbml
# if both of these fail, libsbml cannot be imported - cannot continue
try:
    import tesbml as libsbml
except ImportError:
    import libsbml

from math import isnan
from re import sub

__version__ = "1.2.2"


class sbmlModel:
    """Class to generate sbml model file using libsbml method."""

    def check(self, value, message):
        """Return value to string using libsbml."""

        if value is None:
            raise SystemExit("LibSBML returned a null value trying to " + message + ".")
        if isinstance(value, int):
            if value == libsbml.LIBSBML_OPERATION_SUCCESS:
                return
            err_msg = (
                "Error trying to "
                + message
                + "."
                + "LibSBML returned error code "
                + str(value)
                + ': "'
                + libsbml.OperationReturnValue_toString(value).strip()
                + '"'
            )
            raise RuntimeError(err_msg)
        return

    def __init__(
        self,
        time_units="second",
        extent_units="mole",
        sub_units="mole",
        level=3,
        version=1,
    ):
        if level == 1:
            raise SystemExit("Error: SimpleSBML does not support SBML level 1")
        try:
            self.document = libsbml.SBMLDocument(level, version)
        except ValueError:
            raise SystemExit("Could not create SBMLDocument object")
        self.model = self.document.createModel()
        self.check(self.model, "create model")
        if self.document.getLevel() == 3:
            self.check(self.model.setTimeUnits(time_units), "set model-wide time units")
            self.check(
                self.model.setExtentUnits(extent_units), "set model units of extent"
            )
            self.check(
                self.model.setSubstanceUnits(sub_units), "set model substance units"
            )

        per_second = self.model.createUnitDefinition()
        self.check(per_second, "create unit definition")
        self.check(per_second.setId("per_second"), "set unit definition id")
        unit = per_second.createUnit()
        self.check(unit, "create unit on per_second")
        self.check(unit.setKind(libsbml.UNIT_KIND_SECOND), "set unit kind")
        self.check(unit.setExponent(-1), "set unit exponent")
        self.check(unit.setScale(0), "set unit scale")
        self.check(unit.setMultiplier(1), "set unit multiplier")

        per_min = self.model.createUnitDefinition()
        self.check(per_min, "create unit definition")
        self.check(per_min.setId("per_min"), "set unit definition id")
        unit = per_min.createUnit()
        self.check(unit, "create unit on per_min")
        self.check(unit.setKind(libsbml.UNIT_KIND_SECOND), "set unit kind")
        self.check(unit.setExponent(-1), "set unit exponent")
        self.check(unit.setScale(0), "set unit scale")
        self.check(unit.setMultiplier(60), "set unit multiplier")

        molperLmin = self.model.createUnitDefinition()
        self.check(molperLmin, "create unit definition")
        self.check(molperLmin.setId("molperLmin"), "set unit definition id")
        unit = molperLmin.createUnit()
        self.check(unit, "create unit on per_second")
        self.check(unit.setKind(libsbml.UNIT_KIND_MOLE), "set unit kind")
        self.check(unit.setExponent(1), "set unit exponent")
        self.check(unit.setScale(0), "set unit scale")
        self.check(unit.setMultiplier(1), "set unit multiplier")
        unit = molperLmin.createUnit()
        self.check(unit, "create unit on per_second")
        self.check(unit.setKind(libsbml.UNIT_KIND_LITRE), "set unit kind")
        self.check(unit.setExponent(-1), "set unit exponent")
        self.check(unit.setScale(0), "set unit scale")
        self.check(unit.setMultiplier(1), "set unit multiplier")
        unit = molperLmin.createUnit()
        self.check(unit, "create unit on per_second")
        self.check(unit.setKind(libsbml.UNIT_KIND_SECOND), "set unit kind")
        self.check(unit.setExponent(-1), "set unit exponent")
        self.check(unit.setScale(0), "set unit scale")
        self.check(unit.setMultiplier(60), "set unit multiplier")

        molperL = self.model.createUnitDefinition()
        self.check(molperL, "create unit definition")
        self.check(molperL.setId("molperL"), "set unit definition id")
        unit = molperL.createUnit()
        self.check(unit, "create unit")
        self.check(unit.setKind(libsbml.UNIT_KIND_MOLE), "set unit kind")
        self.check(unit.setExponent(1), "set unit exponent")
        self.check(unit.setScale(0), "set unit scale")
        self.check(unit.setMultiplier(1), "set unit multiplier")
        unit = molperL.createUnit()
        self.check(unit, "create unit")
        self.check(unit.setKind(libsbml.UNIT_KIND_LITRE), "set unit kind")
        self.check(unit.setExponent(-1), "set unit exponent")
        self.check(unit.setScale(0), "set unit scale")
        self.check(unit.setMultiplier(1), "set unit multiplier")

        Dimension_less = self.model.createUnitDefinition()
        self.check(Dimension_less, "create unit definition")
        self.check(Dimension_less.setId("Dimension_less"), "set unit definition id")
        unit = Dimension_less.createUnit()
        self.check(unit, "create unit")
        self.check(unit.setKind(libsbml.UNIT_KIND_DIMENSIONLESS), "set unit kind")
        self.check(unit.setExponent(0), "set unit exponent")
        self.check(unit.setScale(0), "set unit scale")
        self.check(unit.setMultiplier(0), "set unit multiplier")

        self.addCompartment()

    def addCompartment(self, vol=1, comp_id=""):
        """Create compartment of volume litres to the model.

        :param vol: volume of compartment in L
        :type vol: float
        :param comp_id: compartment id
        :type comp_id: str
        """

        c1 = self.model.createCompartment()
        self.check(c1, "create compartment")
        if len(comp_id) == 0:
            comp_id = "c" + str(self.model.getNumCompartments())
        self.check(c1.setId(comp_id), "set compartment id")
        self.check(c1.setConstant(True), 'set compartment "constant"')
        self.check(c1.setSpatialDimensions(3), "set compartment dimensions")

        self.check(c1.setSize(vol), 'set compartment "size"')
        self.check(c1.setUnits("litre"), "set compartment size units")
        return c1

    def addSpecies(self, species_id, amt, comp="c1"):
        """Create Species with the provided amount.

        :param species_id: id or name of the species.
        :type species_id: str
        :param amt: initial amount.
        :type amt: float
        :param comp: compartment id
        :type comp: str
        """

        s1 = self.model.createSpecies()
        self.check(s1, "create species s1")
        self.check(s1.setCompartment(comp), "set species s1 compartment")
        if species_id[0] == "[" and species_id[len(species_id) - 1] == "]":
            self.check(
                s1.setInitialConcentration(amt), "set initial concentration for s1"
            )
            species_id = species_id[1 : (len(species_id) - 1)]
        else:
            self.check(s1.setInitialAmount(amt), "set initial amount for s1")
        self.check(
            s1.setSubstanceUnits(self.model.getSubstanceUnits()),
            "set substance units for s1",
        )
        if species_id[0] == "$":
            self.check(s1.setBoundaryCondition(True), 'set "boundaryCondition" on s1')
            self.check(s1.setConstant(False), 'set "constant" attribute on s1')
            self.check(s1.setId(species_id[1 : len(species_id)]), "set species s1 id")
        else:
            self.check(s1.setBoundaryCondition(False), 'set "boundaryCondition" on s1')
            self.check(s1.setConstant(False), 'set "constant" attribute on s1')
            self.check(s1.setId(species_id), "set species s1 id")
        self.check(
            s1.setHasOnlySubstanceUnits(False), 'set "hasOnlySubstanceUnits" on s1'
        )
        return s1

    def addParameter(self, param_id, val, units="per_second"):
        """Add Parameter with value and unit.

        :param param_id: parameter id/name
        :type param_id: str
        :param val: value for the parameter
        :type val: float
        :param units: unit for the parameter
        :type units: str
        """

        k = self.model.createParameter()
        self.check(k, "create parameter k")
        self.check(k.setId(param_id), "set parameter k id")
        self.check(k.setConstant(True), 'set parameter k "not constant"')
        self.check(k.setValue(val), "set parameter k value")
        self.check(k.setUnits(units), "set parameter k units")
        return k

    def addReaction(
        self, reactants, products, expression, local_params=None, rxn_id=""
    ):
        """Create reaction provided with reactants and products in lists

        :param reactants: list of species id for reactants
        :type reactants: list
        :param products: list of species id for products
        :type products: list
        :param expression: reaction rate expression
        :type expression: str
        :param local_params: keys are the param id and values are their respective values
        :type local_params: dict
        :param rxn_id: id for the reaction
        :type rxn_id: str, optional
        """

        r1 = self.model.createReaction()
        self.check(r1, "create reaction")
        if len(rxn_id) == 0:
            rxn_id = "v" + str(self.model.getNumReactions())
        self.check(r1.setId(rxn_id), "set reaction id")
        self.check(r1.setReversible(False), "set reaction reversibility flag")
        self.check(r1.setFast(False), 'set reaction "fast" attribute')

        for re in reactants:
            if re is not None and "$" in re:
                re.translate(None, "$")
            re_split = re.split()
            if len(re_split) == 1:
                sto = 1.0
                re_id = re
            elif len(re_split) == 2 and re_split[0].isdigit():
                sto = float(re_split[0])
                re_id = re_split[1]
            else:
                err_msg = (
                    "Error: reactants must be listed in format 'S' or '(float)' S'"
                )
                raise SystemExit(err_msg)
            s1 = self.model.getSpecies(re_id)
            species_ref1 = r1.createReactant()
            self.check(species_ref1, "create reactant")
            self.check(species_ref1.setSpecies(s1.getId()), "assign reactant species")
            self.check(
                species_ref1.setStoichiometry(sto), "assign reactant stoichiometry"
            )
            if self.document.getLevel() == 3:
                self.check(
                    species_ref1.setConstant(True), 'set "constant" on species ref 1'
                )

        for pro in products:
            if pro is not None and "$" in pro:
                pro.translate(None, "$")
            pro_split = pro.split()
            if len(pro_split) == 1:
                sto = 1.0
                pro_id = pro
            elif len(pro_split) == 2:
                sto = float(pro_split[0])
                pro_id = pro_split[1]
            else:
                err_msg = "Error: products must be listed in format 'S' or '(float)' S'"
                raise SystemExit(err_msg)
            s2 = self.model.getSpecies(pro_id)
            species_ref2 = r1.createProduct()
            self.check(species_ref2, "create product")
            self.check(species_ref2.setSpecies(s2.getId()), "assign product species")
            self.check(species_ref2.setStoichiometry(sto), "set product stoichiometry")
            if self.document.getLevel() == 3:
                self.check(
                    species_ref2.setConstant(True), 'set "constant" on species ref 2'
                )

        math_ast = libsbml.parseL3Formula(expression)
        self.check(math_ast, "create AST for rate expression")

        kinetic_law = r1.createKineticLaw()
        self.check(kinetic_law, "create kinetic law")
        self.check(kinetic_law.setMath(math_ast), "set math on kinetic law")
        if local_params is not None:
            for param in local_params.keys():
                val = local_params.get(param)
                if self.document.getLevel() == 3:
                    p = kinetic_law.createLocalParameter()
                else:
                    p = kinetic_law.createParameter()
                self.check(p, "create local parameter")
                self.check(p.setId(param), "set id of local parameter")
                self.check(p.setValue(val), "set value of local parameter")
        return r1

    def addEvent(
        self,
        trigger,
        assignments,
        persistent=True,
        initial_value=False,
        priority=0,
        delay=0,
        event_id="",
    ):
        """Add event supplied with when an event is triggered and
        what happens using assignments in a dictionary.

        :param trigger: define when an event is triggered (logical expression)
        :type trigger: str
        :param assignments: keys are the variables to be modified and the values are the new values
        :type assignments: dict
        :param persistent: determine if the event will still be executed if trigger turns from
            True to False
        :type persistent: boolean
        :param initial_value: value of trigger before t=0
        :type initial_value: boolean
        :param priority: determine which event is executed, event with larger priority is executed
        :type priority: float
        :param delay: time between when the event is triggered and the assignment is implemented
        :type delay: float
        :param event_id: id of the event
        :type event_id: str
        """

        e1 = self.model.createEvent()
        self.check(e1, "create event")
        if len(event_id) == 0:
            event_id = "e" + str(self.model.getNumEvents())
        self.check(e1.setId(event_id), "add id to event")
        if self.document.getLevel() == 3 or (
            self.document.getLevel() == 2 and self.document.getVersion() == 4
        ):
            self.check(
                e1.setUseValuesFromTriggerTime(True), "set use values from trigger time"
            )

        tri = e1.createTrigger()
        self.check(tri, "add trigger to event")
        tri_ast = libsbml.parseL3Formula(trigger)
        self.check(tri.setMath(tri_ast), "add formula to trigger")
        if self.document.getLevel() == 3:
            self.check(tri.setPersistent(persistent), "set persistence of trigger")
            self.check(
                tri.setInitialValue(initial_value), "set initial value of trigger"
            )

        de = e1.createDelay()
        if self.document.getLevel() == 3:
            k = self.addParameter(event_id + "Delay", delay, self.model.getTimeUnits())
        else:
            k = self.addParameter(event_id + "Delay", delay, "time")
        self.check(de, "add delay to event")
        delay_ast = libsbml.parseL3Formula(k.getId())
        self.check(de.setMath(delay_ast), "set formula for delay")

        for a in assignments.keys():
            assign = e1.createEventAssignment()
            self.check(assign, "add event assignment to event")
            self.check(assign.setVariable(a), "add variable to event assignment")
            val_ast = libsbml.parseL3Formula(assignments.get(a))
            self.check(assign.setMath(val_ast), "add value to event assignment")

        if self.document.getLevel() == 3:
            pri = e1.createPriority()
            pri_ast = libsbml.parseL3Formula(str(priority))
            self.check(pri.setMath(pri_ast), "add priority to event")
        return e1

    def addAssignmentRule(self, var, math):
        """To assign a state variable with an expression.

        :param var: id of the state variable
        :type var: str
        :param math: expression in str
        :type math: str
        """

        r = self.model.createAssignmentRule()
        self.check(r, "create assignment rule r")
        self.check(r.setVariable(var), "set assignment rule variable")
        math_ast = libsbml.parseL3Formula(math)
        self.check(r.setMath(math_ast), "set assignment rule equation")
        return r

    def addRateRule(self, var, math, rr_id=""):
        """Describe the derivative of the state variable wrt time as an expression.

        :param var: id of the state variable
        :type var: str
        :param math: expression in str
        :type math: str
        :param rr_id: id for the reaction rate
        :type rr_id: str, optional
        """

        r = self.model.createRateRule()
        self.check(r, "create rate rule r")
        if len(rr_id) == 0:
            rr_id = "Rate" + str(self.model.getNumRules())
        r.setIdAttribute(rr_id)
        # print("rr_id=", r.getIdAttribute())

        # self.check(r.setId(rr_id),           'set rate rule id')
        self.check(r.setVariable(var), "set rate rule variable")
        math_ast = libsbml.parseL3Formula(math)
        self.check(r.setMath(math_ast), "set rate rule equation")
        return r

    def addInitialAssignment(self, symbol, math):
        """Describe the initial value of the variable in terms of
        other variables or parameters.

        :param symbol: id of the variable
        :type symbol: str
        :param math: expression
        :type math: str
        """

        if self.document.getLevel() == 2 and self.document.getVersion() == 1:
            raise SystemExit(
                "Error: InitialAssignment does not exist for \
                    this level and version."
            )
        a = self.model.createInitialAssignment()
        self.check(a, "create initial assignment a")
        self.check(a.setSymbol(symbol), "set initial assignment a symbol")
        math_ast = libsbml.parseL3Formula(math)
        self.check(a.setMath(math_ast), "set initial assignment a math")
        return a

    def setLevelAndVersion(self, level, version):
        """Set the level and version of the SBML.

        :param level: level of the sbml
        :type level: int
        :param version: version of the sbml
        :type version: int
        """

        if level == 2 and version == 1:
            self.check(
                self.document.checkL2v1Compatibility(), "convert to level 2 version 1"
            )
        elif level == 2 and version == 2:
            self.check(
                self.document.checkL2v2Compatibility(), "convert to level 2 version 2"
            )
        elif level == 2 and version == 3:
            self.check(
                self.document.checkL2v3Compatibility(), "convert to level 2 version 3"
            )
        elif level == 2 and version == 4:
            self.check(
                self.document.checkL2v4Compatibility(), "convert to level 2 version 4"
            )
        elif level == 3 and version == 1:
            self.check(
                self.document.checkL3v1Compatibility(), "convert to level 3 version 1"
            )
        else:
            raise SystemExit("Invalid level/version combination")

        isSet = self.document.setLevelAndVersion(level, version)
        self.check(isSet, "convert to level " + str(level) + " version " + str(version))

    def getDocument(self):
        """Return document."""
        return self.document

    def getModel(self):
        """Return Model."""
        return self.model

    def getSpecies(self, species_id):
        """Return Species."""
        return self.model.getSpecies(species_id)

    def getListOfSpecies(self):
        """Return list of species."""
        return self.model.getListOfSpecies()

    def getParameter(self, param_id):
        """Return parameter."""
        return self.model.getParameter(param_id)

    def getListOfParameters(self):
        """Return list of parameters."""
        return self.model.getListOfParameters()

    def getReaction(self, rxn_id):
        """Return Reaction."""
        return self.model.getReaction(rxn_id)

    def getListOfReactions(self):
        """Return list of reactions."""
        return self.model.getListOfReactions()

    def getCompartment(self, comp_id):
        """Return compartment."""
        return self.model.getCompartment(comp_id)

    def getListOfEvents(self):
        """Return list of events."""
        return self.model.getListOfEvents()

    def getEvent(self, event_id):
        """Return event."""
        return self.model.getEvent(event_id)

    def getListOfCompartments(self):
        """Return list of compartments."""
        return self.model.getListOfCompartments()

    def getRule(self, var):
        """Return rule."""
        return self.model.getRule(var)

    def getListOfRules(self):
        """Return list of rules."""
        return self.model.getListOfRules()

    def getInitialAssignment(self, var):
        """Return initial assignment."""
        return self.model.getInitialAssignment(var)

    def getListOfInitialAssignments(self):
        """Return list of initial assignments."""
        return self.model.getListOfInitialAssignments()

    def toSBML(self):
        """Return the model in SBML format as strings."""
        errors = self.document.checkConsistency()
        if errors > 0:
            for i in range(errors):
                print(
                    self.document.getError(i).getSeverityAsString(),
                    ": ",
                    self.document.getError(i).getMessage(),
                )

        return libsbml.writeSBMLToString(self.document)

    def __repr__(self):
        return self.toSBML()


def writeCode(doc):
    """Return string containing calls to functions that reproduce the model
    in SBML doc.
    """

    comp_template = "model.addCompartment(vol=%s, comp_id='%s');"
    species_template = "model.addSpecies(species_id='%s', amt=%s, comp='%s');"
    param_template = "model.addParameter(param_id='%s', val=%s, units='%s');"
    rxn_template = (
        "model.addReaction(reactants=%s, products=%s, "
        "expression='%s', local_params=%s, rxn_id='%s');"
    )
    event_template = (
        "model.addEvent(trigger='%s', assignments=%s, persistent=%s, "
        "initial_value=%s, priority=%s, delay=%s, event_id='%s');"
    )
    event_defaults = [True, False, "0", 0]
    assignrule_template = "model.addAssignmentRule(var='%s', math='%s');"
    raterule_template = "model.addRateRule(var='%s', math='%s', rr_id='%s');"
    initassign_template = "model.addInitialAssignment(symbol='%s', math='%s')"
    init_template = (
        "import simplesbml\nmodel = simplesbml.sbmlModel(time_units='%s', "
        "extent_units='%s', sub_units='%s', level=%s, version=%s);"
    )
    init_defaults = ["min", "Molar", "Molar", 3, 1]
    command_list = []

    if doc.getLevel() == 1:
        warnings.warn("Warning: SimpleSBML does not support SBML Level 1.")

    props = libsbml.ConversionProperties()
    props.addOption("flatten comp", True)
    result = doc.convert(props)
    if result != libsbml.LIBSBML_OPERATION_SUCCESS:
        raise SystemExit("Conversion failed: (" + str(result) + ")")

    mod = doc.getModel()
    comps = mod.getListOfCompartments()
    species = mod.getListOfSpecies()
    params = mod.getListOfParameters()
    rxns = mod.getListOfReactions()
    events = mod.getListOfEvents()
    rules = mod.getListOfRules()
    print("rules", rules)
    inits = []
    if doc.getLevel() == 3 or (doc.getLevel() == 2 and doc.getVersion() > 1):
        inits = mod.getListOfInitialAssignments()

    timeUnits = "min"  # second
    substanceUnits = "Molar"  # mole
    extentUnits = "Molar"  # mole
    if doc.getLevel() == 3:
        timeUnits = mod.getTimeUnits()
        extentUnits = mod.getExtentUnits()
        substanceUnits = mod.getSubstanceUnits()
    level = mod.getLevel()
    version = mod.getVersion()
    init_list = [timeUnits, extentUnits, substanceUnits, level, version]
    for i in range(0, 5):
        if init_list[i] == init_defaults[i]:
            init_list[i] = "del"

    command_list.append(
        init_template
        % (init_list[0], init_list[1], init_list[2], init_list[3], init_list[4])
    )

    for comp in comps:
        if comp.getId() != "c1":
            if comp.getId()[0] == "c" and comp.getId()[1 : len(comp.getId())].isdigit():
                if comp.getSize() == 1e-15:
                    command_list.append(comp_template % ("del", "del"))
                else:
                    command_list.append(comp_template % (comp.getSize(), "del"))
            else:
                if comp.getSize() == 1e-15:
                    command_list.append(comp_template % ("del", comp.getId()))
                else:
                    command_list.append(comp_template % (comp.getSize(), comp.getId()))

    for s in species:
        conc = s.getInitialConcentration()
        amt = s.getInitialAmount()
        sid = s.getId()
        if s.getCompartment() == "c1":
            comp = "del"
        else:
            comp = s.getCompartment()
        bc = s.getBoundaryCondition()
        if bc:
            sid = "$" + sid
        if isnan(conc) or amt > conc:
            command_list.append(species_template % (sid, str(amt), comp))
        else:
            command_list.append(species_template % ("[" + sid + "]", str(conc), comp))

    for p in params:
        val = p.getValue()
        pid = p.getId()
        if p.getUnits() == "per_second":
            units = "del"
        else:
            units = p.getUnits()
        isDelay = pid.find("Delay")
        if isDelay == -1:
            command_list.append(param_template % (pid, str(val), str(units)))

    for v in rxns:
        vid = v.getId()
        if vid[0] == "v" and vid[1 : len(vid)].isdigit():
            vid = "del"
        reactants = []
        for r in v.getListOfReactants():
            reactants.append(
                (str(r.getStoichiometry()) + " " + r.getSpecies()).replace("1.0 ", "")
            )
        products = []
        for p in v.getListOfProducts():
            products.append(
                (str(p.getStoichiometry()) + " " + p.getSpecies()).replace("1.0 ", "")
            )
        expr = libsbml.formulaToString(v.getKineticLaw().getMath())
        local_params = {}
        local_ids = []
        local_values = []
        for k in v.getKineticLaw().getListOfParameters():
            local_ids.append(k.getId())
            local_values.append(k.getValue())
        local_params = dict(zip(local_ids, local_values))
        if len(local_params) == 0:
            local_params = "del"
        command_list.append(
            rxn_template % (str(reactants), str(products), expr, str(local_params), vid)
        )

    for e in events:
        persistent = True
        initialValue = False
        priority = "0"
        eid = e.getId()
        if len(eid) == 0 or (eid[0] == "e" and eid[1 : len(eid)].isdigit()):
            eid = "del"
        if doc.getLevel() == 3:
            persistent = e.getTrigger().getPersistent()
            initialValue = e.getTrigger().getInitialValue()
            priority = e.getPriority()
            if isinstance(priority, libsbml.Priority):
                priority = libsbml.formulaToL3String(priority.getMath())
            else:
                priority = "0"
        tri = libsbml.formulaToL3String(e.getTrigger().getMath())
        did = e.getDelay()
        if isinstance(did, libsbml.Delay):
            delay = libsbml.formulaToL3String(did.getMath())
        else:
            delay = "0"
        assigns = e.getListOfEventAssignments()
        var = []
        values = []
        for assign in assigns:
            var.append(assign.getVariable())
            values.append(libsbml.formulaToL3String(assign.getMath()))
        assigns = dict(zip(var, values))

        event_list = [persistent, initialValue, priority, delay]
        for i in range(0, 4):
            if event_list[i] == event_defaults[i]:
                event_list[i] = "del"

        command_list.append(
            event_template
            % (
                tri,
                str(assigns),
                event_list[0],
                event_list[1],
                event_list[2],
                event_list[3],
                eid,
            )
        )

    for r in rules:
        rid = r.getId()
        print("rid")
        #        if rid[0] == 'Rate' and rid[1:len(rid)].isdigit():
        #            rid = 'del'
        sym = r.getVariable()
        math = libsbml.formulaToL3String(r.getMath())
        if r.getTypeCode() == libsbml.SBML_ASSIGNMENT_RULE:
            command_list.append(assignrule_template % (sym, math))
        elif r.getTypeCode() == libsbml.SBML_RATE_RULE:
            command_list.append(raterule_template % (sym, math, rid))
        else:
            pass

    for i in inits:
        sym = i.getSymbol()
        math = libsbml.formulaToL3String(i.getMath())
        command_list.append(initassign_template % (sym, math))

    commands = "\n".join(command_list)
    commands = sub(r"\w+='?del'?(?=[,)])", "", commands)
    commands = sub(r"\((, )+", "(", commands)
    commands = sub(r"(, )+\)", ")", commands)
    commands = sub("(, )+", ", ", commands)
    return commands


def writeCodeFromFile(filename):
    """Read the SBML format model and returns strings containing calls to
    functions to reproduce the model in an sbmlModel object.
    """

    reader = libsbml.SBMLReader()
    doc = reader.readSBMLFromFile(filename)
    if doc.getNumErrors() > 0:
        raise SystemExit(doc.getError(0))
    return writeCode(doc)


def writeCodeFromString(sbmlstring):
    """Read sbmlstring as the SBML format model and return strings containing
    calls to functions to reproduce the model in an sbmlModel object.
    """

    reader = libsbml.SBMLReader()
    doc = reader.readSBMLFromString(sbmlstring)
    return writeCode(doc)



# pylint: disable=C0330,C0103,R0913
"""This module contains a generic class for a well."""
from synbiopython.lab_automation.containers.WellContent import WellContent
from synbiopython.lab_automation.picklist.Transfer import TransferError
from ..tools import unit_factors


class Well:
    """Generic class for a well.

    :param plate: The plate on which the well is located
    :param row: The well's row (a number, starting from 0)
    :param column: The well's column (a number, starting from 0)
    :param name: The well's name, for instance "A1"
    :param data: A dictionary storing data on the well, used in algorithms and reports.
    """

    capacity = None
    dead_volume_per_transfer_class = None

    def __init__(self, plate, row, column, name, data=None):
        self.plate = plate
        self.row = row
        self.column = column
        self.name = name
        self.data = data or {}
        self.sources = []
        self.content = WellContent()

    @property
    def volume(self):
        """Return volume."""
        return self.content.volume

    def iterate_sources_tree(self):
        """Iterate through the tree of sources."""
        for source in self.sources:
            if isinstance(source, Well):
                for parent in source.iterate_sources_tree():
                    yield parent
            else:
                yield source
        yield self

    def add_content(self, components_quantities, volume=None, unit_volume="L"):
        """Add content to well.

        :param components_quantities: Dictionary of components and quantities
          (default: gram). Example `{"Compound_1": 5}`.
        :param volume: Volume (default: liter).
        :param unit_volume: Unit of volume (default: liter). Options: liter (L),
            milliliter (mL), microliter (uL), nanoliter (nL).
        """
        volume = volume * unit_factors[unit_volume]
        if volume > 0:
            final_volume = self.content.volume + volume
            if (self.capacity is not None) and (final_volume > self.capacity):
                raise TransferError(
                    "Transfer of %.2e L to %s brings volume over capacity."
                    % (volume, self)
                )
            self.content.volume = final_volume
        for component, quantity in components_quantities.items():
            if component not in self.content.quantities:
                self.content.quantities[component] = 0
            self.content.quantities[component] += quantity

    def subtract_content(self, components_quantities, volume=0):
        """Subtract content from well."""
        if volume > 0:
            if volume > self.volume:
                raise TransferError(
                    (
                        "Subtraction of %.2e L from %s is impossible."
                        " Current volume: %.2e L"
                    )
                    % (volume, self, self.volume)
                )
            self.content.volume -= volume
        for component, quantity in components_quantities.items():
            if self.content.quantities[component] == quantity:
                self.content.quantities.pop(component)
            else:
                self.content.quantities[component] -= quantity

    def empty_completely(self):
        """Empty the well."""
        self.content.quantities = {}
        self.content.volume = 0

    @property
    def coordinates(self):
        """Return (well.row, well.column)."""
        return (self.row, self.column)

    @property
    def is_empty(self):
        """Return true if the well's volume is 0."""
        return self.volume == 0

    def __repr__(self):
        return "(%s-%s)" % (self.plate.name, self.name)

    def pretty_summary(self):
        """Return a summary string of the well."""
        data = "\n    ".join(
            [""] + [("%s: %s" % (key, value)) for key, value in self.data.items()]
        )
        content = "\n    ".join(
            [""]
            + [
                ("%s: %s" % (key, value))
                for key, value in self.content.quantities.items()
            ]
        )
        return (
            "{self}\n"
            "  Volume: {self.volume}\n"
            "  Content: {content}\n"
            "  Metadata: {data}"
        ).format(self=self, content=content, data=data)

    def to_dict(self):
        """Convert well to dict"""
        return dict(
            [
                ["name", self.name],
                ["content", self.content.to_dict()],
                ["row", self.row],
                ["column", self.column],
            ]
            + list(self.data.items())
        )

    def index_in_plate(self, direction="row"):
        """Return the index of the well in the plate."""
        return self.plate.wellname_to_index(self.name, direction=direction)

    def is_after(self, other, direction="row"):
        """Return whether this well is located strictly after the other well.

        Example: iterate over all free wells after the last non-free well:

        >>> direction = 'row'
        >>> last_occupied_well = plate.last_nonempty_well(direction=direction)
        >>> free_wells = (w for w in plate.iter_wells(direction=direction)
        >>>               if w.is_after(last_occupied_well))
        >>> for well in free_wells: ...
        """
        well_index = self.index_in_plate(direction=direction)
        other_index = other.index_in_plate(direction=direction)
        return well_index > other_index

    def __lt__(self, other):
        return str(self) < str(other)



# pylint: disable=E0401,C0103,R0912,R0914,R0915,R1702
"""
Synbiopython (c) Global BioFoundry Alliance 2020

Synbiopython is licensed under the MIT License.

This module is to implement the simple plotting of the gene circuit using the modified code
from quickplot.py in DNAplotlib library
# added quickplot style for writing Regulations.
# added new Regulation type: Derepression

Reference:
    https://github.com/VoigtLab/dnaplotlib

Install:
    pip install dnaplotlib
"""

import re
import matplotlib
import matplotlib.pyplot as plt
import dnaplotlib as dpl


class SimpleDNAplot:
    """Class to generate SBOL visual compliant gene circuit diagram.
    Regulation type: Connection, Activation, Repression, Derepression
    Each of the part will be numbered sequentially based on the part type/module
    from left to right starting from index 0.
    Example: p0-r0-c0-t0-p1-r1-c1-t1 for two modules with promoter, RBS, coding
    region, and terminator.
    Input = "p.pTet r.rbs34 c.orange.LacI t p.pLac r.rbs32 c.green.TetR t"
    Regulations = "c0->p1.Repression c1->p0.Repression"
    # The default color is black if color is not specified
    """

    def set_circuit_design(self, Input, Regulation=None):
        """Generate the dictionary list containing circuit design information.

        :param Input: a string containing the individual type of part, followed by color
            and name separated by a space.
        :type Input: str
        :param Regulation: a string containing the from_part to the to_part connected
            by an arrow. Type of interaction is specified after the topart followed
            by the color. Default color of black is used is not specified.
        :type Regulation: str, optional
        :return: The part information and Regulations stored in the form of list of dictionaries.
        :rtype: list of dict
        """

        # Types mapping
        types = {}
        types["p"] = "Promoter"
        types["i"] = "Ribozyme"
        types["r"] = "RBS"
        types["c"] = "CDS"
        types["t"] = "Terminator"
        types["s"] = "Spacer"
        types["="] = "Scar"
        types["o"] = "Origin"
        types["es"] = "EmptySpace"

        # Colours mapping
        colors = {}
        colors["black"] = (0.00, 0.00, 0.00)
        colors["gray"] = (0.60, 0.60, 0.60)
        colors["red"] = (0.89, 0.10, 0.11)
        colors["orange"] = (1.00, 0.50, 0.00)
        colors["yellow"] = (1.00, 1.00, 0.00)
        colors["white"] = (1.00, 1.00, 1.00)
        colors["green"] = (0.20, 0.63, 0.17)
        colors["blue"] = (0.12, 0.47, 0.71)
        colors["purple"] = (0.42, 0.24, 0.60)
        colors["lightred"] = (0.98, 0.60, 0.60)
        colors["lightorange"] = (0.99, 0.75, 0.44)
        colors["lightyellow"] = (1.00, 1.00, 0.60)
        colors["lightgreen"] = (0.70, 0.87, 0.54)
        colors["lightblue"] = (0.65, 0.81, 0.89)
        colors["lightpurple"] = (0.79, 0.70, 0.84)

        # Generate the parts list from the arguments
        part_list = []
        part_length = []
        part_idx = 1
        for el in Input.split(" "):
            if el != "":
                part_parts = el.split(".")
                part_label = ""
                part_rgb = (0, 0, 0)
                if len(part_parts) >= 1:
                    part_short_type = part_parts[0]
                    part_fwd = True
                    if part_short_type[0] == "-":
                        part_fwd = False
                        part_short_type = el[1]
                    if part_short_type in types.keys():
                        part_type = types[part_short_type]

                    else:
                        print("Error! Please specify the correct part type!")

                    part_label_yoffset = -5
                    part_label_color = "black"
                    part_label_size = 9
                    part_label_style = "normal"
                    if part_parts[0][0] == "-":
                        part_label_yoffset = 5

                    if len(part_parts) >= 2:
                        part_color = part_parts[1]
                        if part_color in colors.keys():
                            part_rgb = colors[part_color]
                        else:
                            part_label = part_parts[1]

                        if len(part_parts) >= 3:
                            part_label = part_parts[2]
                        elif part_label == "":
                            part_label = ""

                        if part_short_type == "c":
                            part_label_yoffset = 0
                            part_label_color = "white"
                            part_label_size = 10
                            part_label_style = "italic"
                        elif part_short_type == "o":
                            part_label_yoffset = -10
                            part_label_color = "black"
                            part_label_size = 8
                            part_label_style = "normal"
                        elif part_short_type in ("p", "r"):
                            part_label_yoffset = -5
                            part_label_color = "black"
                            part_label_size = 8
                            part_label_style = "normal"
                        else:
                            part_label_yoffset = -5
                            part_label_color = "black"
                            part_label_size = 8
                            part_label_style = "normal"
                            if part_parts[0][0] == "-":
                                part_label_yoffset = 5

                    part_list.append(
                        {
                            "name": str(part_idx),
                            "type": part_type,
                            "fwd": part_fwd,
                            "opts": {
                                "color": part_rgb,
                                "label": part_label,
                                "label_y_offset": part_label_yoffset,
                                "label_color": part_label_color,
                                "label_size": part_label_size,
                                "label_style": part_label_style,
                            },
                        }
                    )
                    part_length.append(part_short_type)

            part_idx = part_idx + 1

        # update the part_length to include the numbering to be used for regulations
        for i in types:
            n = 0
            for a, _ in enumerate(part_length):
                if part_length[a] is i:
                    part_length[a] = i + str(n)
                    n += 1

        print("part_length", part_length)

        # Update the Regulations from the arguments
        Regulations = []

        if Regulation is not None:
            Reg_list = Regulation.split(" ")

            for r, _ in enumerate(Reg_list):
                first_yoffset = 0
                second_yoffset = 0
                arc_height_const = 17
                arc_height_spacing = 4
                arc_height_start = 13

                if Reg_list[r] != "":
                    reg_parts = Reg_list[r].split(".")
                    if "Derepression" in reg_parts[1]:
                        first_yoffset = -5
                        second_yoffset = 4.5
                        reg_type = "Repression"
                    elif "Activation2" in reg_parts[1]:
                        first_yoffset = -5
                        second_yoffset = 4.5
                        reg_type = "Activation"
                    else:
                        second_yoffset = 0
                        reg_type = reg_parts[1]
                    tofr_part = reg_parts[0].split("->")
                    fr_part = tofr_part[0]
                    to_part = tofr_part[1]
                    fr_part_len = self.compute_dnalength(fr_part, part_length)
                    to_part_len = self.compute_dnalength(to_part, part_length)

                    fwd = (fr_part_len <= to_part_len) or (
                        "Derepression" in reg_parts[1]
                    )

                    if len(reg_parts) > 2:
                        reg_color = reg_parts[2]

                        if reg_color in colors.keys():
                            reg_rgb = colors[reg_color]
                    else:
                        reg_rgb = colors["black"]

                    arc_height_spacing = arc_height_spacing + first_yoffset
                    arc_height_const = arc_height_const + second_yoffset
                    arc_height_start = arc_height_start + second_yoffset
                    arc_height_end = arc_height_start * 1
                    arc_height = arc_height_const + arc_height_spacing

                    Regulations.append(
                        {
                            "type": reg_type,
                            "from_part": {"start": fr_part_len, "end": fr_part_len},
                            "to_part": {
                                "start": to_part_len,
                                "end": to_part_len,
                                "fwd": fwd,
                            },
                            "opts": {
                                "color": reg_rgb,
                                "linewidth": 1.5,
                                "arc_height_const": arc_height_const,
                                "arc_height_spacing": arc_height_spacing,
                                "arc_height_start": arc_height_start,
                                "arc_height_end": arc_height_end,
                                "arc_height": arc_height,
                                "arrowhead_length": 2,
                            },
                        }
                    )

            # Modify the plot of the last Repression before Derepression
            indices = [i for i, s in enumerate(Reg_list) if "Derepression" in s]
            indices2 = [i for i, s in enumerate(Reg_list) if "Activation2" in s]

            Rep = []
            for i, _ in enumerate(indices):
                ind = []
                for j in range(indices[i]):
                    if "Repression" in Reg_list[j]:
                        ind.append(j)
                Rep.append({str(indices[i]): ind})

            Act = []
            for i, _ in enumerate(indices2):
                ind = []
                for j in range(indices2[i]):
                    if "Activation" in Reg_list[j]:
                        ind.append(j)
                Act.append({str(indices2[i]): ind})

            arc_height_const = 17
            arc_height_spacing = 4
            first_yoffset = -5
            if len(Rep) != 0:
                for i in Rep:
                    arc_height_spacing = arc_height_spacing + first_yoffset
                    arc_height = arc_height_const + arc_height_spacing
                    Regulations[list(i.values())[0][-1]]["opts"][
                        "arc_height_spacing"
                    ] = arc_height_spacing
                    Regulations[list(i.values())[0][-1]]["opts"][
                        "arc_height"
                    ] = arc_height

            arc_height_const = 17
            arc_height_spacing = 4
            first_yoffset = -5
            if len(Act) != 0:
                for i in Act:
                    arc_height_spacing = arc_height_spacing + first_yoffset
                    arc_height = arc_height_const + arc_height_spacing
                    Regulations[list(i.values())[0][-1]]["opts"][
                        "arc_height_spacing"
                    ] = arc_height_spacing
                    Regulations[list(i.values())[0][-1]]["opts"][
                        "arc_height"
                    ] = arc_height

        else:
            Regulations = None

        return part_list, Regulations

    @staticmethod
    def compute_dnalength(part, part_length):
        """Calculate the position for the to_part or from_part for plotting arrows
        automatically.

        :param part: the to_part or from_part
        :type part: str
        :param part_length: all the parts with sequential numbering starting from 0
        :type part_length: list of str
        :return: dna length
        :rtype: float
        """

        # dna length
        dnalen = {}
        dnalen["p"] = 14.0
        dnalen["i"] = 9.0
        dnalen["r"] = 14.0
        dnalen["c"] = 32.0
        dnalen["t"] = 12.0
        dnalen["s"] = 10.0
        dnalen["="] = 7.0
        dnalen["o"] = 17.0
        dnalen["es"] = 20.0

        dnalength = 0
        ind = part_length.index(part)

        for i in range(ind + 1):
            p = re.sub(r"\d", "", part_length[i])
            if p in dnalen.keys():
                if i is ind:
                    dnalength += 0.5 * dnalen[p]

                else:
                    dnalength += dnalen[p]

        return dnalength

    def plot_circuit(self, Input, Regulation=None, savefig=None):
        """Plot the SBOL-compliant gene circuit figure.

        :param Input: Input design from users
        :type Input: str
        :param Regulation: Regulation strings from users
        :type Regulation: str
        :param savefig: path to store the output figure
        :type savefig: str, optional
        :return: max dna design length and export the gene circuit figure
        :rtype: float
        """

        # matplotlib.use("Qt5Agg")
        matplotlib.use("Agg")

        dr = dpl.DNARenderer(linewidth=1.5)

        # Process the arguments
        design, Regulations = self.set_circuit_design(Input, Regulation)

        reg_renderers = dr.std_reg_renderers()
        part_renderers = dr.SBOL_part_renderers()

        # Generate the figure
        fig = plt.figure(figsize=(1.0, 1.0), dpi=100)
        ax = fig.add_subplot(1, 1, 1)

        # Plot the design
        dna_start, dna_end = dr.renderDNA(
            ax, design, part_renderers, Regulations, reg_renderers
        )
        max_dna_len = dna_end - dna_start

        # print("Max Dna length: ", max_dna_len)

        # Format the axis
        ax.set_xticks([])
        ax.set_yticks([])

        # Set bounds
        ax.set_xlim([(-0.0 * max_dna_len), max_dna_len + (0.0 * max_dna_len)])
        ax.set_ylim([-25, 25])
        ax.set_aspect("equal")
        ax.set_axis_off()

        # Update the size of the figure to fit the constructs drawn
        fig_x_dim = max_dna_len / 30
        # print("x_dim: ", fig_x_dim)
        if fig_x_dim < 1.0:
            fig_x_dim = 1.0
        fig_y_dim = 1.8
        plt.gcf().set_size_inches(fig_x_dim, fig_y_dim, forward=True)

        # Save the figure
        plt.tight_layout()

        if savefig is not None:
            fig.savefig(savefig, transparent=True, dpi=300)
        # plt.show()

        return max_dna_len, fig



# pylint: disable=C0116,C0103,E0401,W0631
"""Miscellaneous useful functions.

In particular, methods for converting to and from plate coordinates.
"""

import numpy as np


def round_at(value, rounding=None):
    """Round value at the nearest rounding.

    :param value: the value to round
    """
    if rounding is None:
        return value
    return np.round(value / rounding) * rounding


def replace_nans_in_dict(dictionary, replace_by="null"):
    """Replace NaNs in a dictionary with a string.

    :param dictionary: the dictionary
    :type dictionary: dict
    :param replace_by: replacement
    :type replace_by: str
    """
    for key, value in dictionary.items():
        if isinstance(value, dict):
            replace_nans_in_dict(value, replace_by=replace_by)
        elif value is np.nan:
            dictionary[key] = replace_by


def human_seq_size(n):
    "Return the given sequence as a human friendly 35b, 1.4k, 15k, etc."
    if n < 1000:
        return "%db" % n
    if n < 10000:
        return "%.1fk" % (n / 1000)

    return "%dk" % np.round(n / 1000)


unit_factors = {
    prefix + unit: factor
    for unit in "glL"
    for prefix, factor in [("", 1), ("m", 1e-3), ("u", 1e-6), ("n", 1e-9)]
}

volume_values_and_units = sorted(
    [(value, unit) for (unit, value) in unit_factors.items() if unit.endswith("L")]
)


def find_best_volume_unit(vols):
    """Find the best volume unit for a list of volumes."""
    med = np.median(vols)
    for value, unit in volume_values_and_units:
        if (not unit.endswith("g")) and (med <= 999 * value):
            return unit
    return unit


def human_volume(vol, unit="auto"):
    """Return a human-readable volume."""
    if unit == "auto":
        unit = find_best_volume_unit([vol])
    vol = np.round(vol / unit_factors[unit], 2)
    if int(vol) == vol:
        return "%d %s" % (vol, unit)

    return "%s %s" % (("%.02f" % vol).rstrip("0"), unit)



# pylint: disable=C0114,C0103,C0116
import math
import re


def compute_rows_columns(num_wells):
    """Convert 96->(8,12), 384->(16,24), etc."""
    a = math.sqrt(num_wells / 6)
    n_rows = int(round(2 * a))
    n_columns = int(round(3 * a))
    return n_rows, n_columns


def rowname_to_number(name):
    "Convert A->1 Z->26 AA->27 etc."
    if len(name) == 2:
        return 26 * rowname_to_number(name[0]) + rowname_to_number(name[1])
    try:
        return "ABCDEFGHIJKLMNOPQRSTUVWXYZ".index(name) + 1
    except IndexError:
        raise ValueError(name + " is not a valid row name.")


def number_to_rowname(number):
    "Convert 1->A 26->Z 27->AA etc."
    if number > 26:
        return number_to_rowname(int(number / 26)) + number_to_rowname(number % 26)
    return "ABCDEFGHIJKLMNOPQRSTUVWXYZ"[number - 1]


def wellname_to_coordinates(wellname):
    """Convert A1->(1,1), H11->(8, 11), etc."""
    rowname, colname = re.match("([a-zA-Z]+)([0-9]+)", wellname).groups()
    return rowname_to_number(rowname), int(colname)


def coordinates_to_wellname(coords):
    """Convert (1,1)->A1, (4,3)->D3, (12, 12)->H12, etc."""
    row, column = coords
    return number_to_rowname(row) + str(column)


def wellname_to_index(wellname, num_wells, direction="row"):
    """Convert e.g. A1..H12 into 1..96
    direction is either row for A1 A2 A3... or column for A1 B1 C1 D1 etc.

    :param wellname: the name of the well
    :param num_wells: number of wells on the plate
    :type num_wells: int
    :param direction: the direction of counting. Either "row" or "column".
    :type direction: str
    """
    n_rows, n_columns = compute_rows_columns(num_wells)
    row, column = wellname_to_coordinates(wellname)
    if direction == "row":
        return column + n_columns * (row - 1)
    if direction == "column":
        return row + n_rows * (column - 1)
    raise ValueError("`direction` must be in (row, column)")


def index_to_row_column(index, num_wells, direction="row"):
    n_rows, n_columns = compute_rows_columns(num_wells)
    if direction == "row":
        row = 1 + int((index - 1) / n_columns)
        column = 1 + ((index - 1) % n_columns)
    elif direction == "column":
        row, column = 1 + ((index - 1) % n_rows), 1 + int((index - 1) / n_rows)
    else:
        raise ValueError("`direction` must be in (row, column)")
    return row, column


def index_to_wellname(index, num_wells, direction="row"):
    """Convert e.g. 1..96 into A1..H12

    :param index: the index of the well
    :type index: int
    :param num_wells: number of wells on the plate
    :type num_wells: int
    :param direction: the direction of counting. Either "row" or "column".
    :type direction: str
    """
    row, column = index_to_row_column(index, num_wells, direction)
    return coordinates_to_wellname((row, column))



"""
Synbiopython (c) Global BioFoundry Alliance 2020

Synbiopython is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author: neilswainston
"""
import os.path

import pandas as pd
from synbiopython.codon import DATA_DIR


def _get_spec_df():
    """Get NCBI Taxonomy ids and organism from species.table as a Pandas
    DataFrame.

    :return: a Pandas DataFrame, containing NCBI Taxonomy ids and organism
        names.
    :rtype: pd.DataFrame
    """
    spec_tab_df = pd.read_csv(
        os.path.join(DATA_DIR, "species.table"),
        names=["name", "tax_id"],
        sep="\t",
        comment="#",
    )

    return spec_tab_df.set_index("tax_id")


_SPEC_DF = _get_spec_df()


def get_tax_id(table_id):
    """Gets a NCBI Taxonomy id from supplied parameter, which may be either an
    organism name or a NCBI Taxonomy id.

    :param table_id: an organism name or a NCBI Taxonomy id (as either a str or
        int).
    :type table_id: str
    :return: a NCBI Taxonomy id
    :rtype: str
    """
    table_id = str(table_id)

    if table_id in _SPEC_DF.index:
        return table_id

    tax_ids = _SPEC_DF.index[_SPEC_DF["name"] == table_id]

    if tax_ids.any():
        return tax_ids[0]

    raise ValueError("Unrecognised table id: %s" % table_id)


def get_organism_name(table_id):
    """Gets an organism name from supplied parameter, which may be either an
    organism name or a NCBI Taxonomy id.

    :param table_id: an organism name or a NCBI Taxonomy id (as either a str or
        int).
    :type table_id: str
    :return: an organism name
    :rtype: str
    """
    tax_id = get_tax_id(table_id)
    return _SPEC_DF.loc[tax_id, "name"] if tax_id else None



# pylint: disable=C0103,E0401
"""
Synbiopython (c) Global BioFoundry Alliance 2020

Synbiopython is licensed under the MIT License.

This module provides code to work with SBOL validator
https://validator.sbolstandard.org/

Use sample sbol file from github (can also be obtained from iBioSim)

Reference:
    https://github.com/SynBioDex/SBOL-Validator/blob/master/src/test/sequence1.xml
    http://synbiodex.github.io/SBOL-Validator/#query-parameters
    http://synbiodex.github.io/SBOL-Validator/#options
    http://biopython.org/DIST/docs/tutorial/Tutorial.html (Chapter 17 Graphics)

install:
    pip install biopython reportlab

The URI prefix is required for FASTA and GenBank conversion,
and optional for SBOL 1 conversion
"""

import os
import requests
from Bio import SeqIO
from Bio.Graphics import GenomeDiagram
from reportlab.lib import colors
from reportlab.lib.units import cm


class GenSBOLconv:
    """Class to convert standard files (SBOL1, SBOL2, GenBank, Fasta, GFF3)."""

    @staticmethod
    def export_plasmidmap(gbfile, filename=None):
        """ Export Linear and Circular Plasmid Map for the imported GenBank file.

        :param gbfile: a genbank file in .gb format or the path the file if not in the
            same folder.
        :type gbfile: str
        :param filename: the filenames/path to the filenames for the linear and
            circular plasmids in tuple
        :type filename: tuple, optional
        :return: the version from the genbank file
        :rtype: str
        """

        record = SeqIO.read(gbfile, "genbank")

        gd_diagram = GenomeDiagram.Diagram(record.id)
        gd_track_for_features = gd_diagram.new_track(1, name="Annotated Features")
        gd_feature_set = gd_track_for_features.new_set()

        for feature in record.features:
            if feature.type == "primer" or (feature.type == "misc_feature"):
                continue
            #            if (feature.type != "CDS"):
            #                # Exclude this feature
            #                continue
            if len(gd_feature_set) % 2 == 0:
                color = colors.lightblue

            else:
                color = colors.blue

            gd_feature_set.add_feature(
                feature,
                sigil="ARROW",
                color=color,
                label_size=12,
                label_angle=0,
                label=True,
            )

        # Draw Linear map from genbank
        gd_diagram.draw(
            format="linear",
            orientation="landscape",
            pagesize="A4",
            fragments=4,
            start=0,
            end=len(record),
        )
        if filename is None:
            linfile = "plasmid_linear.png"
            circfile = "plasmid_circular.png"
        else:
            linfile = filename[0]
            circfile = filename[1]

        gd_diagram.write(linfile, "PNG")

        # Draw circular map from genbank
        gd_diagram.draw(
            format="circular",
            circular=True,
            pagesize=(25 * cm, 20 * cm),  # pagesize=(35 * cm, 30 * cm),
            start=0,
            end=len(record),
            circle_core=0.5,
        )
        # gd_diagram.write("plasmid_circular.pdf", "PDF")
        gd_diagram.write(circfile, "PNG")

        return record.id

    @staticmethod
    def access_sbolvalidator(input_file, Output, uri_Prefix=""):
        """Code to invoke the SBOL Validator server over the internet.

        :param input_file: input filename or filepath
        :type input_file: str
        :param Output: the type of Output file
        :type Output: str, ('GenBank', 'FASTA', 'GFF3', 'SBOL1', 'SBOL2')
        :param uri_Prefix: '' as default, URI Prefix is required for FASTA and GenBank
            input conversion
        :type uri_Prefix: str, optional
        :return: POST request response from webpage
        :rtype: object
        """

        file = open(input_file).read()

        request = {
            "options": {
                "language": Output,
                "test_equality": False,
                "check_uri_compliance": False,
                "check_completeness": False,
                "check_best_practices": False,
                "fail_on_first_error": False,
                "provide_detailed_stack_trace": False,
                "subset_uri": "",
                "uri_prefix": uri_Prefix,
                "version": "",
                "insert_type": False,
                "main_file_name": "main file",
                "diff_file_name": "comparison file",
            },
            "return_file": True,
            "main_file": file,
        }

        # send POST request to the specified url (Response [200] means ok)
        response = requests.post(
            "https://validator.sbolstandard.org/validate/", json=request
        )

        return response

    @staticmethod
    def get_outputfile_extension(Filetype):
        """Get the output file extension based on the requested output language.

        :param Filetype: the type of Output file
        :type Output: str, ('GenBank', 'FASTA', 'GFF3', 'SBOL1', 'SBOL2')
        :return: the specific file extension
        :rtype: str
        """

        switcher = {
            "GenBank": ".gb",
            "FASTA": ".fasta",
            "GFF3": ".gff",
            "SBOL1": ".sbol",
            "SBOL2": ".sbol",
        }
        return switcher.get(Filetype, "unknown filetype")

    def export_outputfile(self, input_filename, Response, Output, outputfile=None):
        """Export the converted output file.

        :param input_filename: input filename or filepath
        :type input_filename: str
        :param Response: response from POST request to sbolvalidator web page
        :type Response: object
        :param Output: the type of Output file
        :type Output: str, ('GenBank', 'FASTA', 'GFF3', 'SBOL1', 'SBOL2')
        :param outputfile: provide specific outputfilename or filepath
        :type outputfile: str, optional
        """

        filename_w_ext = os.path.basename(input_filename)
        filename, _ = os.path.splitext(filename_w_ext)

        if Response.json()["valid"]:
            # export the result from json into the specific output file format
            if outputfile is None:
                output_filename = filename + self.get_outputfile_extension(Output)
            else:
                output_filename = outputfile

            print("Output file: ", output_filename)

            with open(output_filename, "w", newline="\n") as f:
                f.write(Response.json()["result"])
        else:
            print("Error message: ", Response.json()["errors"])

    def run_sbolvalidator(self, Input_file, Output, uri_Prefix="", **kwargs):
        """Wrapper function for the SBOL Validator.

        :param Input_file: input file or path to input file
        :type Input_file: str, filename or filepath
        :param Output: the type of Output file
        :type Output: str, ('GenBank', 'FASTA', 'GFF3', 'SBOL1', 'SBOL2')
        :param uri_Prefix: '' as default, URI Prefix is required for FASTA and GenBank
            input conversion
        :type uri_Prefix: str, optional
        :return: the validity of the Response, and export output file.
        :rtype: str, "valid: True" if the conversion is done properly
        :Keyword Arguments:
            * *outputfile*: specify outputfile
        """
        Response = self.access_sbolvalidator(Input_file, Output, uri_Prefix)

        output_filename = None

        for key, value in kwargs.items():
            if "outputfile" in key:
                output_filename = value

        self.export_outputfile(Input_file, Response, Output, outputfile=output_filename)

        return "valid: " + str(Response.json()["valid"])



# pylint: disable=C0330,C0103,E1101,R0913,E0102,R1705,E0401
"""This module implements the Base class for all plates.

See synbiopython.lab_automation.containers for more specific plate subclasses, with
set number of wells, well format, etc.
"""
from collections import OrderedDict
import pandas
from synbiopython.lab_automation.containers.Well import Well
from synbiopython.lab_automation.containers.helper_functions import (
    index_to_wellname,
    number_to_rowname,
    wellname_to_index,
    coordinates_to_wellname,
    rowname_to_number,
)
from synbiopython.lab_automation.tools import replace_nans_in_dict


class NoUniqueWell(Exception):
    """NoUniqueWell exception class."""


class Plate:
    """Base class for all plates.

    See the builtin_containers for usage classes, such as generic microplate
    classes (Plate96, Plate384, etc).

    :param name: Name or ID of the Plate as it will appear in strings and reports
    :param wells_data: A dict {"A1": {data}, "A2": ...}.
        The format of the data is left free
    :param plate_data: plate data
    """

    well_class = Well

    def __init__(self, name=None, wells_data=None, plate_data=None):

        self.name = name
        self.data = plate_data or {}
        self.wells_data = wells_data or {}
        self.num_wells = self.num_rows * self.num_columns
        self.wells = {}
        self.columns = {column: [] for column in range(1, self.num_columns + 1)}
        self.rows = {number_to_rowname(row): [] for row in range(1, self.num_rows + 1)}
        for row in range(1, self.num_rows + 1):
            for column in range(1, self.num_columns + 1):
                wellname = coordinates_to_wellname((row, column))
                data = self.wells_data.get(wellname, {})
                well = self.well_class(
                    plate=self,
                    row=row,
                    column=column,
                    name=wellname,
                    data=data,
                )
                self.wells[wellname] = well
                self.columns[column] += [wellname]
                self.rows[number_to_rowname(row)] += [wellname]

    def __getitem__(self, k):
        """Return e.g. well A1's dict when calling `myplate['A1']`."""
        return self.wells[k]

    def find_unique_well_by_condition(self, condition):
        """Return the unique well of the plate satisfying the condition.

        The ``condition`` method should have a signature of Well=>True/False.

        Raises a NoUniqueWell error if 0 or several wells satisfy the condition.
        """
        wells = [well for name, well in self.wells.items() if condition(well)]
        if len(wells) > 1:
            raise NoUniqueWell("Query returned several wells: %s" % wells)
        if len(wells) == 0:
            raise NoUniqueWell("No wells found matching the condition")
        return wells[0]

    def find_unique_well_containing(self, query):
        """Return the unique well whose content contains the query."""

        def condition(well):
            return query in well.content.quantities.keys()

        return self.find_unique_well_by_condition(condition)

    def list_well_data_fields(self):
        """Return all fields used in well data in the plate."""
        return sorted(list(set(field for well in self for field in well.data.keys())))

    def return_column(self, column_number):
        """Return the list of all wells of the plate in the given column."""
        return [self.wells[wellname] for wellname in self.columns[column_number]]

    def list_wells_in_column(self, column_number):
        """Return the list of all wells of the plate in the given column.

        Examples:

        >>> for well in plate.list_wells_in_column(5):
        >>>      print(well.name)
        """
        return [well for well in self.iter_wells() if well.column == column_number]

    def return_row(self, row):
        """Return the list of all wells of the plate in the given row.

        The `row` can be either a row number (1,2,3) or row letter(s) (A,B,C).
        """
        if isinstance(row, int):
            row = number_to_rowname(row)
        return [self.wells[wellname] for wellname in self.rows[row]]

    def list_wells_in_row(self, row):
        """Return the list of all wells of the plate in the given row.

        The `row` can be either a row number (1,2,3) or row letter(s) (A,B,C).

        Examples:

        >>> for well in plate.list_wells_in_row("H"):
        >>>      print(well.name)

        """
        if isinstance(row, str):
            row = rowname_to_number(row)
        return [well for well in self.iter_wells() if well.row == row]

    def list_filtered_wells(self, well_filter):
        """List filtered wells.

        Examples:

        >>> def condition(well):
        >>>     return well.volume > 50
        >>> for well in myplate.list_filtered_wells(condition):
        >>>     print(well.name)
        """
        return list(filter(well_filter, self.wells.values()))

    def wells_grouped_by(
        self,
        data_field=None,
        key=None,
        sort_keys=False,
        ignore_none=False,
        direction_of_occurence="row",
    ):
        """Return wells grouped by key."""
        if key is None:

            def key(well):
                return well.data.get(data_field, None)

        dct = OrderedDict()
        for well in self.iter_wells(direction=direction_of_occurence):
            well_key = key(well)
            if well_key not in dct:
                dct[well_key] = [well]
            else:
                dct[well_key].append(well)
        if ignore_none:
            dct.pop(None, None)
        keys = dct.keys()
        if sort_keys:
            keys = sorted(keys)
        return [(k, dct[k]) for k in keys]

    def get_well_at_index(self, index, direction="row"):
        """Return the well at the corresponding index.

        Examples:

        >>> plate.get_well_at_index(1)  # well A1
        >>> plate.get_well_at_index(2)  # well A2
        >>> plate.get_well_at_index(2, direction="column")  # well B1
        """
        return self[self.index_to_wellname(index, direction=direction)]

    def index_to_wellname(self, index, direction="row"):
        """Return the name of the well at the corresponding index.

        Examples:

        >>> plate.index_to_wellname(1)  # "A1"
        >>> plate.get_well_at_index(2)  # "A2"
        >>> plate.get_well_at_index(2, direction="column")  # "B1"
        """
        return index_to_wellname(index, self.num_wells, direction=direction)

    def wellname_to_index(self, wellname, direction="row"):
        """Return the index of the well in the plate.

        Examples:
        >>> plate.wellname_to_index("A1")  # 1
        >>> plate.wellname_to_index("A2")  # 2
        >>> plate.wellname_to_index("A1", direction="column")  # 9 (8x12 plate)
        """
        return wellname_to_index(wellname, self.num_wells, direction=direction)

    def wells_sorted_by(self, sortkey):
        """Return wells sorted by sortkey"""
        return (e for e in sorted(self.wells.values(), key=sortkey))

    def iter_wells(self, direction="row"):
        """Iter through the wells either by row or by column.

        Examples:

        >>> for well in plate.iter_wells():
        >>>     print (well.name)
        """
        if direction == "row":
            return self.wells_sorted_by(lambda w: (w.row, w.column))
        else:
            return self.wells_sorted_by(lambda w: (w.column, w.row))

    def to_dict(self, replace_nans_by="null"):
        """Convert plate to dict."""
        dct = {
            "data": self.data,
            "wells": {well.name: well.to_dict() for well in self.wells.values()},
        }
        if replace_nans_by is not None:
            replace_nans_in_dict(dct, replace_by=replace_nans_by)
        return dct

    def to_pandas_dataframe(self, fields=None, direction="row"):
        """Return a dataframe with the info on each well."""
        dataframe = pandas.DataFrame.from_records(self.to_dict()["wells"]).T
        by = ["row", "column"] if direction == "row" else ["column", "row"]
        dataframe = dataframe.sort_values(by=by)
        if fields is not None:
            dataframe = dataframe[fields]
        return dataframe

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name)



# pylint: disable=C0330,C0103,E0102,R1705,R0913
"""Classes to represent picklists and liquid transfers in general."""
from copy import deepcopy
from synbiopython.lab_automation.picklist.Transfer import Transfer


class PickList:
    """Representation of a list of well-to-well transfers.

    :param transfers_list: A list of Transfer objects that will be part of the same
        dispensing operation, in the order in which they are meant to be simulated.
    :param data: A dict with information on the picklist.
    """

    def __init__(self, transfers_list=(), data=None):

        self.transfers_list = list(transfers_list)
        self.data = {} if data is None else data

    def add_transfer(
        self,
        source_well=None,
        destination_well=None,
        volume=None,
        data=None,
        transfer=None,
    ):
        """Add a transfer to the picklist's tranfers list.

        You can either provide a ``Transfer`` object with the ``transfer``
        parameter, or the parameters.
        """
        if transfer is None:
            transfer = Transfer(
                source_well=source_well,
                destination_well=destination_well,
                volume=volume,
                data=data,
            )
        self.transfers_list.append(transfer)

    def to_plain_string(self):
        """Return the list of transfers in human-readable format."""
        return "\n".join(transfer.to_plain_string() for transfer in self.transfers_list)

    def to_plain_textfile(self, filename):
        """Write the picklist in a file in a human reable format."""
        with open(filename, "w+") as f:
            f.write(self.to_plain_string())

    def simulate(self, content_field="content", inplace=True):
        """Simulate the execution of the picklist."""

        if not inplace:
            all_plates = set(
                plate
                for transfer in self.transfers_list
                for plate in [
                    transfer.source_well.plate,
                    transfer.destination_well.plate,
                ]
            )
            new_plates = {plate: deepcopy(plate) for plate in all_plates}

            new_transfer_list = []
            for transfer in self.transfers_list:
                new_source_plate = new_plates[transfer.source_well.plate]
                new_dest_plate = new_plates[transfer.destination_well.plate]
                new_source_well = new_source_plate.wells[transfer.source_well.name]
                new_dest_well = new_dest_plate.wells[transfer.destination_well.name]
                new_transfer_list.append(
                    Transfer(
                        volume=transfer.volume,
                        source_well=new_source_well,
                        destination_well=new_dest_well,
                    )
                )

            new_picklist = PickList(transfers_list=new_transfer_list)
            new_picklist.simulate(
                content_field=content_field,
                inplace=True,
            )
            return new_plates

        else:
            for transfer in self.transfers_list:
                transfer.apply()
            return None

    def restricted_to(
        self, transfer_filter=None, source_well=None, destination_well=None
    ):
        """Return a version of the picklist restricted to transfers with the
        right source/destination well.

        You can provide ``source_well`` and ``destination_well`` or
        alternatively just a function ``transfer_filter`` with signature
        (transfer)=>True/False that will be used to filter out transfers
        (for which it returns false).
        """
        if transfer_filter is None:

            def transfer_filter(tr):
                source_well_is_ok = (source_well is None) or (
                    source_well == tr.source_well
                )
                dest_well_is_ok = (destination_well is None) or (
                    destination_well == tr.destination_well
                )
                return source_well_is_ok and dest_well_is_ok

        transfers = [tr for tr in self.transfers_list if transfer_filter(tr)]
        return PickList(transfers, data={"parent": self})

    def sorted_by(self, sorting_method="source_well"):
        """Return a new version of the picklist sorted by some parameter.

        The ``sorting_method`` is either the name of an attribute of the
        transfers, such as "source_well", or a function f(transfer) -> value.
        """
        if not hasattr(sorting_method, "__call__"):

            def sorting_method(transfer):
                return transfer.__dict__[sorting_method]

        return PickList(
            sorted(self.transfers_list, key=sorting_method),
            data={"parent": self},
        )

    def total_transferred_volume(self):
        """Return the sum of all volumes from all transfers."""
        return sum([transfer.volume for transfer in self.transfers_list])

    def enforce_maximum_dispense_volume(self, max_dispense_volume):
        """Return a new picklist were every too-large dispense is broken down
        into smaller dispenses."""
        transfers = []
        for trf in self.transfers_list:
            n_additional_dispense = int(trf.volume / max_dispense_volume)
            rest = trf.volume - n_additional_dispense * max_dispense_volume
            for _ in range(n_additional_dispense):
                transfers.append(trf.with_new_volume(max_dispense_volume))
            if rest > 0:
                transfers.append(trf.with_new_volume(rest))
        return PickList(transfers_list=transfers)

    def __add__(self, other):
        return PickList(self.transfers_list + other.transfers_list)

    @staticmethod
    def merge_picklists(picklists_list):
        """Merge the list of picklists into a single picklist.

        The transfers in the final picklist are the concatenation of the
        transfers in the different picklists, in the order in which they appear
        in the list.
        """
        return sum(picklists_list, PickList([]))



# pylint: disable=C0103
"""This module contains a class to represent the volume and quantities of a well."""


class WellContent:
    """Class to represent the volume and quantities of a well.

    Having the well content represented as a separate object makes it possible
    to have several wells share the same content, e.g. in throughs.
    """

    def __init__(self, quantities=None, volume=0):
        if quantities is None:
            quantities = {}
        self.volume = volume
        self.quantities = quantities

    def concentration(self, component=None, default=0):
        """Return concentration of component."""
        if self.quantities == {}:
            return default
        if self.volume == 0:
            return default
        if component is None:
            component = list(self.quantities.keys())[0]
        if component not in self.quantities:
            return default
        return 1.0 * self.quantities[component] / self.volume

    def to_dict(self):
        """Return a dict {volume: 0.0001, quantities: {...:...}}."""
        return {"volume": self.volume, "quantities": self.quantities}

    def make_empty(self):
        """Empty the well."""
        self.volume = 0
        self.quantities = {}

    def components_as_string(self, separator=" "):
        """Return a string representation of what's in the well mix."""
        return separator.join(sorted(self.quantities.keys()))



# pylint: disable=C0103,C0114, W0122
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = {}
with open("synbiopython/version.py") as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="synbiopython",
    version=version["__version__"],
    author="Global Biofoundries Alliance",
    description="Python tools for Synthetic Biology.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://synbiopython.org",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pandas>=0.22",
        "numpy",
        "dnaplotlib",
        "biopython",
        "reportlab",
        "tesbml",
        "nbgitpuller",
        "tellurium",
    ],
    include_package_data=True,
)



# pylint: disable=C0103,E0402,R0913,R0914,R0903,R0904,R0801
"""
Synbiopython (c) Global BioFoundry Alliance 2020

Synbiopython is licensed under the MIT License.

This module is to create SBML file for ODE model using simplesbml package,
which relies on libSBML.

Reference:
    https://github.com/sys-bio/simplesbml
"""

from synbiopython.genbabel.sbmlgen import simplesbml
from synbiopython.genbabel import utilities


class SBMLgen:
    """Class to generate SBML file for ODE model."""

    @staticmethod
    def export_sbml(ODE, Variable, Init, ParamName, Param, ParamUnit, **kwargs):
        """Function to generate the SBML xml file.

        :param ODE: The ODEs in the form of string stored in a list
        :type ODE: list
        :param Variable: The names of variable in a list of string
        :type Variable: list
        :param Init: Initial conditions for the variables in a list of values
        :type Init: list
        :param ParamName: The names of the parameters stored in a list
        :type ParamName: list
        :param Param: The parameters values
        :type Param: list
        :param ParamUnit: The unit for the parameter according to available unit definition
        :type ParamUnit: list
        :return: SBML in str
        :rtype: str
        """

        for u in ParamUnit:
            if u == "molL-1min-1":
                ParamUnit[ParamUnit.index(u)] = "molperLmin"
            elif u == "molL-1":
                ParamUnit[ParamUnit.index(u)] = "molperL"
            elif u == "s-1":
                ParamUnit[ParamUnit.index(u)] = "per_second"
            elif u == "min-1":
                ParamUnit[ParamUnit.index(u)] = "per_min"
            elif u == "dimensionless":
                ParamUnit[ParamUnit.index(u)] = "Dimension_less"
            else:
                print("Error in the defined units for parameters")

        model = simplesbml.sbmlModel()

        for s, _ in enumerate(Variable):
            model.addSpecies("[" + Variable[s] + "]", Init[s])

        for p, _ in enumerate(Param):
            model.addParameter(ParamName[p], Param[p], ParamUnit[p])

        for r, _ in enumerate(ODE):
            model.addRateRule(Variable[r], ODE[r])

        Model = model.toSBML()
        XMLfilename = "SBML_" + utilities.getfilename() + ".xml"
        output_file = XMLfilename
        for key, value in kwargs.items():
            if "outputfile" in key:
                output_file = value

        print(Model, file=open(output_file, "w"))
        return Model
